# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import (
    Iterator, Optional, Union,
)

from amundsen_common.utils.atlas import (
    AtlasCommonParams, AtlasCommonTypes, AtlasTableTypes,
)
from amundsen_rds.models import RDSModel
from amundsen_rds.models.application import Application as RDSApplication, ApplicationTable as RDSApplicationTable

from databuilder.models.atlas_entity import AtlasEntity
from databuilder.models.atlas_relationship import AtlasRelationship
from databuilder.models.atlas_serializable import AtlasSerializable
from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable
from databuilder.models.table_metadata import TableMetadata
from databuilder.models.table_serializable import TableSerializable
from databuilder.serializers.atlas_serializer import get_entity_attrs
from databuilder.utils.atlas import AtlasRelationshipTypes, AtlasSerializedEntityOperation


class GenericApplication(GraphSerializable, TableSerializable, AtlasSerializable):
    """
    An Application that generates or consumes a resource.
    """

    LABEL = 'Application'
    DEFAULT_KEY_FORMAT = 'application://{application_type}/{application_id}'

    APP_URL = 'application_url'
    APP_NAME = 'name'
    APP_ID = 'id'
    APP_DESCRIPTION = 'description'

    GENERATES_REL_TYPE = 'GENERATES'
    DERIVED_FROM_REL_TYPE = 'DERIVED_FROM'
    CONSUMES_REL_TYPE = 'CONSUMES'
    CONSUMED_BY_REL_TYPE = 'CONSUMED_BY'

    LABELS_PERMITTED_TO_HAVE_USAGE = ['Table']

    def __init__(self,
                 start_label: str,
                 start_key: str,
                 application_type: str,
                 application_id: str,
                 application_url: str,
                 application_description: Optional[str] = None,
                 app_key_override: Optional[str] = None,  # for bw-compatibility only
                 generates_resource: bool = True,
                 ) -> None:

        if start_label not in GenericApplication.LABELS_PERMITTED_TO_HAVE_USAGE:
            raise Exception(f'applications associated with {start_label} are not supported')

        self.start_label = start_label
        self.start_key = start_key
        self.application_type = application_type
        self.application_id = application_id
        self.application_url = application_url
        self.application_description = application_description
        self.application_key = app_key_override or GenericApplication.DEFAULT_KEY_FORMAT.format(
            application_type=self.application_type,
            application_id=self.application_id,
        )
        self.generates_resource = generates_resource

        self._node_iter = self._create_node_iterator()
        self._relation_iter = self._create_relation_iterator()
        self._record_iter = self._create_record_iterator()
        self._atlas_entity_iterator = self._create_next_atlas_entity()
        self._atlas_relation_iterator = self._create_atlas_relation_iterator()

    def create_next_node(self) -> Union[GraphNode, None]:
        # creates new node
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self) -> Union[GraphRelationship, None]:
        try:
            return next(self._relation_iter)
        except StopIteration:
            return None

    def create_next_record(self) -> Union[RDSModel, None]:
        try:
            return next(self._record_iter)
        except StopIteration:
            return None

    def _create_node_iterator(self) -> Iterator[GraphNode]:
        """
        Create an application node
        :return:
        """
        attrs = {
            GenericApplication.APP_NAME: self.application_type,
            GenericApplication.APP_ID: self.application_id,
            GenericApplication.APP_URL: self.application_url,
        }
        if self.application_description:
            attrs[GenericApplication.APP_DESCRIPTION] = self.application_description

        yield GraphNode(
            key=self.application_key,
            label=GenericApplication.LABEL,
            attributes=attrs,
        )

    def _create_relation_iterator(self) -> Iterator[GraphRelationship]:
        """
        Create relations between application and table nodes
        :return:
        """
        graph_relationship = GraphRelationship(
            start_key=self.start_key,
            start_label=self.start_label,
            end_key=self.application_key,
            end_label=GenericApplication.LABEL,
            type=(GenericApplication.DERIVED_FROM_REL_TYPE if self.generates_resource
                  else GenericApplication.CONSUMED_BY_REL_TYPE),
            reverse_type=(GenericApplication.GENERATES_REL_TYPE if self.generates_resource
                          else GenericApplication.CONSUMES_REL_TYPE),
            attributes={}
        )
        yield graph_relationship

    # TODO: support consuming/producing relationships and multiple apps per resource
    def _create_record_iterator(self) -> Iterator[RDSModel]:
        yield RDSApplication(
            rk=self.application_key,
            application_url=self.application_url,
            name=self.application_type,
            id=self.application_id,
            description=self.application_description or '',
        )

        yield RDSApplicationTable(
            rk=self.start_key,
            application_rk=self.application_key,
        )

    def create_next_atlas_entity(self) -> Union[AtlasEntity, None]:
        try:
            return next(self._atlas_entity_iterator)
        except StopIteration:
            return None

    def _create_next_atlas_entity(self) -> Iterator[AtlasEntity]:
        group_attrs_mapping = [
            (AtlasCommonParams.qualified_name, self.application_key),
            ('name', self.application_type),
            ('id', self.application_id),
            ('description', self.application_description or ''),
            ('application_url', self.application_url)
        ]

        entity_attrs = get_entity_attrs(group_attrs_mapping)

        entity = AtlasEntity(
            typeName=AtlasCommonTypes.application,
            operation=AtlasSerializedEntityOperation.CREATE,
            relationships=None,
            attributes=entity_attrs,
        )

        yield entity

    def create_next_atlas_relation(self) -> Union[AtlasRelationship, None]:
        try:
            return next(self._atlas_relation_iterator)
        except StopIteration:
            return None

    # TODO: support consuming/producing relationships and multiple apps per resource
    def _create_atlas_relation_iterator(self) -> Iterator[AtlasRelationship]:
        yield AtlasRelationship(
            relationshipType=AtlasRelationshipTypes.table_application,
            entityType1=AtlasTableTypes.table,
            entityQualifiedName1=self.start_key,
            entityType2=AtlasCommonTypes.application,
            entityQualifiedName2=self.application_key,
            attributes={}
        )


class AirflowApplication(GenericApplication):

    ID_FORMAT = '{dag}/{task}'
    KEY_FORMAT = 'application://{cluster}.airflow/{id}'
    DESCRIPTION_FORMAT = 'Airflow with id {id}'

    def __init__(self,
                 task_id: str,
                 dag_id: str,
                 application_url_template: str,
                 db_name: str = 'hive',
                 cluster: str = 'gold',
                 schema: str = '',
                 table_name: str = '',
                 application_type: str = 'Airflow',
                 exec_date: str = '',
                 generates_table: bool = True,
                 ) -> None:

        self.database = db_name
        self.cluster = cluster
        self.schema = schema
        self.table = table_name
        self.dag = dag_id
        self.task = task_id

        airflow_app_id = AirflowApplication.ID_FORMAT.format(dag=dag_id, task=task_id)
        GenericApplication.__init__(
            self,
            start_label=TableMetadata.TABLE_NODE_LABEL,
            start_key=self.get_table_model_key(),
            application_type=application_type,
            application_id=airflow_app_id,
            application_url=application_url_template.format(dag_id=dag_id),
            application_description=AirflowApplication.DESCRIPTION_FORMAT.format(id=airflow_app_id),
            app_key_override=AirflowApplication.KEY_FORMAT.format(cluster=cluster, id=airflow_app_id),
            generates_resource=generates_table,
        )

    def get_table_model_key(self) -> str:
        return TableMetadata.TABLE_KEY_FORMAT.format(
            db=self.database,
            cluster=self.cluster,
            schema=self.schema,
            tbl=self.table,
        )


# Alias for backwards compatibility
Application = AirflowApplication



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import importlib
import logging
from typing import Any

from pyhocon import ConfigTree

from databuilder.models.table_metadata import TableMetadata
from databuilder.models.type_metadata import ScalarTypeMetadata
from databuilder.transformer.base_transformer import Transformer

PARSING_FUNCTION = 'parsing_function'

LOGGER = logging.getLogger(__name__)


class ComplexTypeTransformer(Transformer):
    """
    Transforms complex types for columns in a table.
    Takes a TableMetadata object and iterates over the columns.
    The configured parser takes each column's type string, name,
    and the column itself, and sets the column's type_metadata
    field with the parsed results contained in a TypeMetadata object.
    """
    def init(self, conf: ConfigTree) -> None:
        self.success_count = 0
        self.failure_count = 0

        parsing_function = conf.get_string(PARSING_FUNCTION)
        module_name, function_name = parsing_function.rsplit(".", 1)
        mod = importlib.import_module(module_name)
        self._parsing_function = getattr(mod, function_name)

    def transform(self, record: Any) -> TableMetadata:
        if not isinstance(record, TableMetadata):
            raise Exception(f"ComplexTypeTransformer expects record of type TableMetadata, received {type(record)}")

        for column in record.columns:
            try:
                column.set_column_key(record._get_col_key(column))
                column.set_type_metadata(self._parsing_function(column.type, column.name, column))
            except Exception as e:
                # Default to scalar type if the type string cannot be parsed
                column.set_type_metadata(ScalarTypeMetadata(name=column.name, parent=column, type_str=column.type))
                self.failure_count += 1
                LOGGER.warning(f"Could not parse type for column {column.name} in table {record.name}: {e}")
            else:
                self.success_count += 1

        return record

    def get_scope(self) -> str:
        return 'transformer.complex_type'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import (
    Any, Callable, Dict, List, Optional, Tuple,
)

from amundsen_gremlin.neptune_bulk_loader import api as neptune_bulk_loader_api
from boto3.session import Session
from gremlin_python.process.graph_traversal import (
    GraphTraversal, GraphTraversalSource, __,
)
from gremlin_python.process.traversal import T
from pyhocon import ConfigFactory, ConfigTree

from databuilder import Scoped


class NeptuneSessionClient(Scoped):
    """
        A convenience client for neptune gives functions to perform upserts, deletions and queries with filters.
    """
    # What property is used to local nodes and edges by ids
    NEPTUNE_HOST_NAME = 'neptune_host_name'
    # AWS Region the Neptune cluster is located
    AWS_REGION = 'aws_region'
    AWS_ACCESS_KEY = 'aws_access_key'
    AWS_SECRET_ACCESS_KEY = 'aws_access_secret'
    AWS_SESSION_TOKEN = 'aws_session_token'

    WEBSOCKET_OPTIONS = 'websocket_options'

    DEFAULT_CONFIG = ConfigFactory.from_dict(
        {
            AWS_SESSION_TOKEN: None,
            WEBSOCKET_OPTIONS: {},
        }
    )

    def __init__(self) -> None:
        self._graph = None

    def init(self, conf: ConfigTree) -> None:
        conf = conf.with_fallback(NeptuneSessionClient.DEFAULT_CONFIG)

        boto_session = Session(
            aws_access_key_id=conf.get_string(NeptuneSessionClient.AWS_ACCESS_KEY, default=None),
            aws_secret_access_key=conf.get_string(NeptuneSessionClient.AWS_SECRET_ACCESS_KEY, default=None),
            aws_session_token=conf.get_string(NeptuneSessionClient.AWS_SESSION_TOKEN, default=None),
            region_name=conf.get_string(NeptuneSessionClient.AWS_REGION, default=None)
        )
        self._neptune_host = conf.get_string(NeptuneSessionClient.NEPTUNE_HOST_NAME)
        neptune_uri = "wss://{host}/gremlin".format(
            host=self._neptune_host
        )
        source_factory = neptune_bulk_loader_api.get_neptune_graph_traversal_source_factory(
            neptune_url=neptune_uri,
            session=boto_session
        )
        self._graph = source_factory()

    def get_scope(self) -> str:
        return 'neptune.client'

    def get_graph(self) -> GraphTraversalSource:
        return self._graph

    def upsert_node(self, node_id: str, node_label: str, node_properties: Dict[str, Any]) -> None:
        create_traversal = __.addV(node_label).property(T.id, node_id)
        node_traversal = self.get_graph().V().has(T.id, node_id). \
            fold().coalesce(__.unfold(), create_traversal)

        node_traversal = NeptuneSessionClient.update_entity_properties_on_traversal(node_traversal, node_properties)
        node_traversal.next()

    def upsert_edge(
            self,
            start_node_id: str,
            end_node_id: str,
            edge_id: str,
            edge_label: str,
            edge_properties: Dict[str, Any]
    ) -> None:
        create_traversal = __.V().has(
            T.id, start_node_id
        ).addE(edge_label).to(__.V().has(T.id, end_node_id)).property(T.id, edge_id)
        edge_traversal = self.get_graph().V().has(T.id, start_node_id).outE(edge_label).has(T.id, edge_id). \
            fold(). \
            coalesce(__.unfold(), create_traversal)

        edge_traversal = NeptuneSessionClient.update_entity_properties_on_traversal(edge_traversal, edge_properties)
        edge_traversal.next()

    @staticmethod
    def update_entity_properties_on_traversal(
            graph_traversal: GraphTraversal,
            properties: Dict[str, Any]
    ) -> GraphTraversal:
        for key, value in properties.items():
            key_split = key.split(':')
            key = key_split[0]
            value_type = key_split[1]
            if "Long" in value_type:
                value = int(value)
            graph_traversal = graph_traversal.property(key, value)

        return graph_traversal

    @staticmethod
    def filter_traversal(
            graph_traversal: GraphTraversal,
            filter_properties: List[Tuple[str, Any, Callable]],
    ) -> GraphTraversal:
        for filter_property in filter_properties:
            (filter_property_name, filter_property_value, filter_operator) = filter_property
            graph_traversal = graph_traversal.has(filter_property_name, filter_operator(filter_property_value))
        return graph_traversal

    def delete_edges(
            self,
            filter_properties: List[Tuple[str, Any, Callable]],
            edge_labels: Optional[List[str]]
    ) -> None:
        tx = self.get_graph().E()
        if edge_labels:
            tx = tx.hasLabel(*edge_labels)
        tx = NeptuneSessionClient.filter_traversal(tx, filter_properties)

        tx.drop().iterate()

    def delete_nodes(
            self,
            filter_properties: List[Tuple[str, Any, Callable]],
            node_labels: Optional[List[str]]
    ) -> None:
        tx = self.get_graph().V()
        if node_labels:
            tx = tx.hasLabel(*node_labels)
        tx = NeptuneSessionClient.filter_traversal(tx, filter_properties)

        tx.drop().iterate()
